"""
File for obstacles, non-controlled objects with prescribed geometry
"""

class CircularObstacle:
    """
    Class for a circular obstacle. The obstacle should not be interfaced with directly!
    Rather, the obstacle interfaces with a depth camera/lidar object.
    """
    def __init__(self, q, r):
        """
        Init function for an obstacle.
        Inputs:
            q (Nx1 NumPy Array): [X, Y] or [X, Y, Z] position of the obstacle in space
            r (float): radius of obstacle
        """
        #store input parameters
        self._q = q.reshape((q.size, 1)) #reshape to correct (Nx1)
        self._r = r

        #store dimension of shape
        self.dimn = (q.shape)[0]

    def get_center(self):
        """
        Returns the center of the obstacle
        """
        return self._q
    
    def get_radius(self):
        """
        Returns the radius of the obstacle
        """
        return self._r
    
    def set_center(self, qNew):
        """
        Reset the center position of the obstacle
        Inputs:
            qNew (Nx1 NumPy Array): new center position for the obstacle
        """
        self._q = qNew.reshape((qNew.size, 1))
    
    def set_radius(self, rNew):
        """
        Reset the radius of the obstacle
        Inputs:
            rNew (float): new radius for the obstacle
        """
        self._r = rNew


class ObstacleManager:
    def __init__(self, qMatrix, rList, NumObs = 1):
        """
        Managerial class for a set of N obstacles
        Inputs:
            qMatrix (N x NumObs NumPy Array): Matrix containing positions of each obstacle
            rList (list): Python 
            NumObs (Int): number of obstacles
        """
        self.qMatrix = qMatrix
        self.rList = rList
        self.NumObs = NumObs

        #create a dictionary storing N obstacle instances
        self.obsDict = {}

        #create N obstacle objects with a position from qTotal
        for i in range(self.NumObs):
            #create an obstacle with center and radius from qMatrix and rList
            self.obsDict[i] = CircularObstacle(self.qMatrix[:, i], self.rList[i])

    def get_obstacle_i(self, i):
        """
        Function to retrieve the ith obstacle object
        """
        return self.obsDict[i]


"""
This file implements the twist class.
"""
import numpy as np
from .transforms import *

class Twist:
    def __init__(self, v, omega):
        """
        Init function for a unit twist
        Inputs:
            v (3x1 NumPy Array): linear component of twist
            omega (3x1 NumPy Array): angular component of twist
        """
        self._v = v.reshape((3, 1))
        self._omega = omega.reshape((3, 1))
        self._xi = np.vstack((self._v, self._omega))

    def get_v(self):
        """
        Returns linear component of twist
        """
        return self._v
    
    def get_omega(self):
        """
        Returns angular component of twist
        """
        return self._omega
    
    def get_xi(self):
        """
        Returns 6x1 (v, omega)
        """
        return self._xi
    
    def set_v(self, v):
        #reset v
        self._v = v.reshape((3, 1))
        #reset xi
        self._xi = np.vstack((self._v, self._omega))
    
    def set_omega(self, omega):
        #reset omega
        self._omega = omega.reshape((3, 1))
        #reset xi
        self._xi = np.vstack((self._v, self._omega))

    def set_xi(self, v, omega):
        self._v = v.reshape((3, 1))
        self._omega = omega.reshape((3, 1))
        self._xi = np.vstack((self._v, self._omega))

    def exp(self, theta):
        """
        Calculate the matrix exponential of the twist with 
        magnitude theta.
        Inputs:
            theta (float): magnitude (linear distance/angle in radians)
        Returns:
            g ((4x4) NumPy Array): SEE(3) transformation
        """
        return exp_transform(self._xi, theta)


import numpy as np
from .controller import ControllerManager, Controller
from .state_estimation import ObserverManager

class Environment:
    def __init__(self, dynamics, controller = None, observer = None, obstacleManager = None, T = 10):
        """
        Initializes a simulation environment
        Args:
            dynamics (Dynamics): system dynamics object
            controller (Controller): system controller object
            observer (Observer): system state estimation object
            obstacleManager (ObstacleManager): obstacle objects
            T (Float): simulation time
        """
        #store system parameters
        self.dynamics = dynamics #store system dynamics
        self.obstacleManager = obstacleManager #store manager for any obstacles present

        #if observer and controller are none, create default objects
        if observer is not None:
            self.observer = observer
        else:
            #create a default noise-free observer manager
            self.observer = ObserverManager(dynamics)
        if controller is not None:
            self.controller = controller
        else:
            #create a default zero input controller manager (using the skeleton Controller class)
            self.controller = ControllerManager(self.observer, None, None, None, Controller)
        
        #define environment parameters
        self.iter = 0 #number of iterations
        self.t = 0 #time in seconds 
        self.done = False
        
        #Store system state
        self.x = self.dynamics.get_state() #Actual state of the system
        self.x0 = self.x #store initial condition for use in reset
        self.xObsv = None #state as read by the observer
        self.ptCloud = None #point cloud state as read by vision
        
        #Define simulation parameters
        self.SIM_FREQ = 1000 #integration frequency in Hz
        self.CONTROL_FREQ = 50 #control frequency in Hz
        self.SIMS_PER_STEP = self.SIM_FREQ//self.CONTROL_FREQ
        self.TOTAL_SIM_TIME = T #total simulation time in s
        self.TOTAL_ITER = self.TOTAL_SIM_TIME*self.CONTROL_FREQ + 1 #total number of iterations
        
        #Define history arrays
        self.xHist = np.zeros((self.dynamics.sysStateDimn, self.TOTAL_ITER))
        self.uHist = np.zeros((self.dynamics.sysInputDimn, self.TOTAL_ITER))
        self.tHist = np.zeros((1, self.TOTAL_ITER))
        
    def reset(self):
        """
        Reset the gym environment to its inital state.
        """
        #Reset gym environment parameters
        self.iter = 0 #number of iterations
        self.t = 0 #time in seconds
        self.done = False
        
        #Reset system state
        self.x = self.x0 #retrieves initial condiiton
        self.xObsv = None #reset observer state
        
        #Define history arrays
        self.xHist = np.zeros((self.dynamics.sysStateDimn, self.TOTAL_ITER))
        self.uHist = np.zeros((self.dynamics.sysInputDimn, self.TOTAL_ITER))
        self.tHist = np.zeros((1, self.TOTAL_ITER))

    def step(self):
        """
        Step the sim environment by one integration
        """
        #retrieve current state information
        self._get_observation() #updates the observer
        
        #solve for the control input using the observed state
        self.controller.eval_input(self.t)

        #update the deterministic system data, iterations, and history array
        self._update_data()
        
        #Zero order hold over the controller frequency and step dynamics
        for i in range(self.SIMS_PER_STEP):
            self.dynamics.integrate(self.controller.get_input(), self.t, 1/self.SIM_FREQ) #integrate dynamics
            self.t += 1/self.SIM_FREQ #increment the time
    
    def _update_data(self):
        """
        Update history arrays and deterministic state data
        """
        #append the input, time, and state to their history queues
        self.xHist[:, self.iter] = self.x.reshape((self.dynamics.sysStateDimn, ))
        self.uHist[:, self.iter] = (self.controller.get_input()).reshape((self.dynamics.sysInputDimn, ))
        self.tHist[:, self.iter] = self.t
        
        #update the actual state of the system
        self.x = self.dynamics.get_state()
        
        #update the number of iterations of the step function
        self.iter +=1
    
    def _get_observation(self):
        """
        Updates self.xObsv using the observer data
        Useful for debugging state information.
        """
        self.xObsv = self.observer.get_state()
        # print("current orientation: ", self.observer.get_orient())
    
    def _get_reward(self):
        """
        Calculate the total reward for ths system and update the reward parameter.
        Only implement for use in reinforcement learning.
        """
        return 0
    
    def _is_done(self):
        """
        Check if the simulation is complete
        Returns:
            boolean: whether or not the time has exceeded the total simulation time
        """
        #check if we have exceeded the total number of iterations
        if self.iter >= self.TOTAL_ITER:
            return True
        return False
    
    def run(self, N = 1, run_vis = True, verbose = False, obsManager = None):
        """
        Function to run the simulation N times
        Inputs:
            N (int): number of simulation examples to run
            run_vis (bool): run the visualization of the results
        Returns:
            self.xHist, self.uHist, self.tHist
        """
        #loop over an overall simulation N times
        for i in range(N):
            self.reset()
            print("Running Simulation.")
            while not self._is_done():
                if verbose:
                    print("Simulation Time Remaining: ", self.TOTAL_SIM_TIME - self.t)
                self.step() #step the environment while not done
            if run_vis:
                self.visualize() #render the result
        #return the history arrays
        return self.xHist, self.uHist, self.tHist
            
    def visualize(self):
        """
        Provide visualization of the environment
        Inputs:
            obsManager (Obstacle Manager, optional): manager to plot obstacles in the animation
        """
        self.dynamics.show_animation(self.xHist, self.uHist, self.tHist, obsManager = self.obstacleManager)
        self.dynamics.show_plots(self.xHist, self.uHist, self.tHist)


from control import *

#tests for the control functions
def peak_time_test():
    #test peak time
    tData = np.array(range(629))
    yData = np.array([np.sin(x) for x in np.arange(0, 2*np.pi, 0.01)])
    Tp = calc_peak_time(yData, tData)
    assert(np.array_equal(Tp, np.array([[157]])))


peak_time_test()
print("Everything passed!")